#! /usr/bin/env python
# -*- coding: utf-8 -*-
import pkg_resources

# interrogate version string of already-installed distribution
__version__ = pkg_resources.require('poppy-core')[0].version

from poppy.core.tools.exceptions import MissingArgument, \
    MissingProperty, TargetFileNotSaved, MissingInput



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import collections

__all__ = ["DotDict", "SimpleDotDict"]


class DotDict(collections.OrderedDict):
    """
    A class to have ordered dictionary and also be able to access parameters
    inside nested dictionary easily through a dot syntax for the key.
    """
    def __getitem__(self, key):
        """
        Override to access recursively the nested dictionary through the
        dot syntax.
        """
        # separator not present, returns the default value of dict
        if "." not in key:
            return super(DotDict, self).__getitem__(key)

        # the separator is present, split the key
        first, rest = key.split(".", 1)

        # get the value
        value = super(DotDict, self).__getitem__(first)

        # check it is an instance of the class
        if not isinstance(value, DotDict):
            raise KeyError("Cannot get {0} in {1}".format(rest, first))

        # now call the get
        return value.__getitem__(rest)

    def __setitem__(self, key, value):
        """
        Override to set recursively the nested dictionary through the dot
        syntax.
        """
        # the dot is present in the key, split
        if "." in key:
            first, rest = key.split(".", 1)

            # get the element of the first key
            element = self[first]

            # check that it is a DotDict
            if not isinstance(element, DotDict):
                raise KeyError("Cannot set {0} in {1}".format(rest, first))

            # set the value
            element[rest] = value
        else:
            if isinstance(value, dict) and not isinstance(value, DotDict):
                for k, v in value.items():
                    self[k] = v
            super(DotDict, self).__setitem__(key, value)


class SimpleDotDict(dict):
    """
    A class to have dictionary and also be able to access parameters inside
    nested dictionary easily through a dot syntax for the key.
    """
    def __getitem__(self, key):
        """
        Override to access recursively the nested dictionary through the
        dot syntax.
        """
        # separator not present, returns the default value of dict
        if "." not in key:
            return super(SimpleDotDict, self).__getitem__(key)

        # the separator is present, split the key
        first, rest = key.split(".", 1)

        # get the value
        value = super(SimpleDotDict, self).__getitem__(first)

        # check it is an instance of the class
        if not isinstance(value, SimpleDotDict):
            raise KeyError("Cannot get {0} in {1}".format(rest, first))

        # now call the get
        return value.__getitem__(rest)

    def __setitem__(self, key, value):
        """
        Override to set recursively the nested dictionary through the dot
        syntax.
        """
        # the dot is present in the key, split
        if "." in key:
            first, rest = key.split(".", 1)

            # get the element of the first key
            element = self[first]

            # check that it is a SimpleDotDict
            if not isinstance(element, SimpleDotDict):
                raise KeyError("Cannot set {0} in {1}".format(rest, first))

            # set the value
            element[rest] = value
        else:
            if isinstance(value, dict) and not isinstance(
                value,
                SimpleDotDict,
            ):
                for k, v in value.items():
                    self[k] = v
            super(SimpleDotDict, self).__setitem__(key, value)


# vim: set tw=79 :



#!/usr/bin/env python3

import sys
import os

if __name__ == "__main__":

    # automatically add plugins from the plugins folder to the installed python package
    # uncomment the following lines to enable this option
    # ROOT_DIRECTORY = os.path.dirname(os.path.abspath(__file__))
    # plugins_path = os.path.join(ROOT_DIRECTORY, 'plugins')
    # for entry in os.listdir(plugins_path):
    #     dir_path = os.path.join(plugins_path, entry)
    #     if os.path.isdir(dir_path):
    #         sys.path.insert(0, dir_path)

    # run the pipeline
    try:
        from poppy.pop.scripts import main
    except ImportError as exc:
        raise ImportError(
            "Couldn't import poppy-pop. Are you sure it's installed and "
            "available on your PYTHONPATH environment variable? Did you "
            "forget to activate a virtual environment?"
        ) from exc


    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import collections
from poppy.core.logger import logger

from poppy.core.generic.signals import Signal
from poppy.core.tools.poppy_argparse import argparse
from poppy.core.generic.metaclasses import ManagerMeta
from poppy.core.pipeline import Pipeline
__all__ = ["Command"]


def after_creation(cls, name):
    """
    Some extra things to do after the creation of command classes (and not
    instances, this is not an error).
    """
    # check that the class has the good attributes
    if not hasattr(cls, "__command__"):
        logger.error(
            "The class {0} has no attribute __command__".format(name)
        )
        return
    if not hasattr(cls, "__parent__"):
        logger.error(
            "The class {0} has no attribute __parent__".format(name)
        )
        return


class CommandManager(object):
    """
    A class to manage the available defined commands by the user through the
    plugin command classes.
    """

    def __init__(self):
        """
        Create containers where to store defined commands.
        """
        # containers
        self.commands = {}
        self.instancesByCommand = {}

        # tree of dependencies for commands
        self.tree = collections.defaultdict(list)

        # define the arguments specific to the pipeline
        self.pipeline_arguments = {
            'settings': (
                ['--settings'],
                {'help': 'The path to the settings file.'},
            ),

            'log_level': (
                ['-ll', '--log-level'],
                {
                    'help': 'Specifies the amount information that the pipeline should print to the console ',
                    'choices': ('DEBUG', 'INFO', 'WARNING', 'ERROR')
                }
            )
        }

        # signals
        self.added = Signal()
        self.created = Signal()
        self.generation_start = Signal()
        self.generation_stop = Signal()
        self.run_start = Signal()
        self.run_end = Signal()

    def add(self, name, cls):
        """
        Adds the command and its class to the manager.
        """
        # register
        logger.debug("Register the command {0}".format(name))
        self.commands[cls.__command__] = cls

        # add its dependence in the tree (none parent is the root tree)
        self.tree[cls.__parent__].append(cls.__command__)

        # send a signal of a new definition
        self.added(name, cls)

    def create(self, instance):
        """
        Register the instances of the commands by their name.
        """
        # register
        logger.debug(
            "Register command instance {0}".format(instance.__command__)
        )
        self.instancesByCommand[instance.__command__] = instance

        # send the signal of a new instantiation
        self.created(instance)

    def generate(self, options):
        """
        Given a first base subparser and the parents parser, generate the
        parsers for children commands recursively.
        """
        # send signal that the generation is starting
        self.generation_start(options)

        # generate commands
        logger.debug("Generating commands hierarchy")
        self._generate()

        # send signal that the generation stopped
        self.generation_stop(options)

    def _generate(self, subparser=None, start=None, parsers=None):
        # keep trace of the parsers generated by commands
        if parsers is None:
            self.parsers = {}

        # loop over children
        for command in self.tree[start]:
            # create an instance of the command
            instance = self.commands[command]()

            # generate the list of parent parsers used by the command if
            # provided by the user
            if hasattr(instance, "__parent_arguments__"):
                parents = [
                    self.parsers[name]
                    for name in instance.__parent_arguments__
                ]
            else:
                parents = []

            # get the parser for the given command
            parser = instance.parser(subparser, parents)

            # add arguments to the parser from the user
            instance.add_arguments(parser)

            # by default run the instance when the command is called
            parser.set_defaults(run=instance)

            # keep trace of the parser
            self.parsers[instance.__command__] = parser

            # if the command has children, create a subparser and recall the
            # method recursively
            if instance.has_children():
                # generate the subcommands for this command
                self._generate(
                    instance.subparser(parser),
                    command,
                    self.parsers,
                )

    def launch(self, argv=None):
        """
        Launch the commands by parsing the input of the program and then
        calling the good command with the good arguments.
        """

        # get arguments
        args, unknown = self.parse(argv)

        if unknown != []:
            logger.warning(f'There may be wrong/unused arguments : {unknown}')

        # a signal/hook to do something before running the command with the
        # parsed arguments
        self.run_start(args)

        # call the good command
        args.run(args)

        # a signal/hook to do something after running the command with the
        # parsed arguments
        self.run_end(args)

    def parse(self, argv=None):
        """
        Responsible to parse the command line, and also check the consistency
        of the tree of commands for the base command.
        """
        # first check that the number of commands linked to the root that does
        # nothing is not greater than one. If it is not the case, several
        # parsers are defined and thus we are unable to define which one to
        # use.
        if len(self.tree[None]) > 1:
            logger.error(
                "Multiple parsers are defined for the commands. Cannot " +
                "decide which one to use."
            )
            raise SystemExit(-1)
        elif len(self.tree[None]) <= 0:
            logger.error("No commands defined for the root command.")
            raise SystemExit(-1)

        # else parse the root parser
        return self.parsers[self.tree[None][0]].parse_known_args(args=argv)

    def preprocess_args(self, argv):
        """
        Preprocess options like --settings and --config.

        These options could affect the pipeline behavior, so they must be processed early.

        :param argv:
        :return: options, args
        """
        logger.debug('Preprocessing of settings, log-level, etc. arguments')

        parser = argparse.ArgumentParser(description='Preprocess some arguments.', add_help=False, allow_abbrev=False)

        for argument in self.pipeline_arguments.values():
            parser.add_argument(*argument[0], **argument[1])

        return parser.parse_known_args(argv)

    def __getitem__(self, name):
        """
        Get instance of command by its name.
        """
        return self.instancesByCommand[name]

    def __contains__(self, name):
        """
        Check if the instance of a command is present in the manager.
        """
        return name in self.instancesByCommand


class CommandMetaclass(ManagerMeta):
    """
    To manage the definition and instantiation of commands.
    """

    def __call__(cls, *args, **kwargs):
        """
        Make the command instances as singletons.
        """
        # check that the instance is already created or not
        if cls.__command__ in cls.manager:
            return cls.manager[cls.__command__]

        # create the instance as usual
        return super(CommandMetaclass, cls).__call__(*args, **kwargs)


class Command(
    object,
    metaclass=CommandMetaclass,
    manager=CommandManager,
    after_creation=after_creation,
):
    """
    Base class for all accepted commands for the pop command line program.
    """

    def parser(self, subparser, parents):
        """
        Return the parser for the command and options that this command must
        use. Take as argument the subparser from the parent parser.
        """
        return subparser.add_parser(
            self.__command_name__,
            help=self.__help__,
            parents=parents,
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
            conflict_handler="resolve",
        )

    def add_arguments(self, parser):
        """
        Used by the user to add arguments associated to the given parser.
        """
        pass

    def subparser(self, parser):
        """
        Should return a subparser for this command. Not always called, just
        when the command has possibly subcommands to generate.
        """
        if hasattr(self, "__command_name__"):
            command = self.__command_name__
        else:
            command = self.__command__

        return parser.add_subparsers(
            help="Sub-commands available for command {0}".format(command),
        )

    def has_children(self):
        """
        To know if the command has subcommands.
        """
        # check that the command has subcommands
        return len(self.manager.tree[self.__command__]) > 0

    def print_help(self):
        """
        Print the command help message
        """

        Command.manager.parsers[self.__command__].print_help()

    def __call__(self, args):
        """
        The command is called with its arguments as parameters, in order to
        have the possibility to run the succession of commands necessary.
        """

        # create a pipeline instance
        pipeline = Pipeline(args)

        try:
            # setup the tasks
            self.setup_tasks(pipeline)

            # run the pipeline
            pipeline.run()
        except NotImplementedError:
            # if no tasks are defined, print the help message
            self.print_help()

    def setup_tasks(self, pipeline):
        """
        Set up the task workflow
        """
        raise NotImplementedError()


    @classmethod
    def add_parent_parser(cls, name, parser):
        """
        Used to add a parser with its options and be able to refer from a
        command, since the conflict handler of argparse is not well done, as
        many other things.
        """
        # check that the parser with this name is not already present
        if name in cls.manager.parsers:
            logger.error(
                "Trying to add parser with name {0} while already defined"
            )
            raise SystemExit(-1)

        # add the parser to the collection
        cls.manager.parsers[name] = parser

    def __repr__(self):
        return "Command {0}".format(self.__command__)

    def __str__(self):
        return self.__command__

# vim: set tw=79 :



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Pipeline Timer
"""

import threading

from poppy.core.generic.signals import Signal

__all__ = ["Timer"]


class Timer(object):
    """
    A class timer to make periodic calls to a given function or single shot
    calls after a given amount of time.
    """
    def __init__(self):
        # the signal to emit when the time ended
        self.timeout = Signal()

        # initiate some variables
        self._timer = None
        self._timerSingle = None

    def _run(self):
        # create a new timer with the specified time
        self._timer = threading.Timer(self._time, self._run)
        self._timer.start()

        # emit the signal to inform connected slots
        self.timeout()

    def singleShot(self, time=None):
        """
        To run the timer just one time after a given delay.
        """
        self._timerSingle = threading.Timer(time, self.timeout)
        self._timerSingle.start()

    def start(self, time):
        """
        To start the periodic timer.
        """
        self._time = time
        self._timer = threading.Timer(self._time, self._run)
        self._timer.start()

    def stop(self):
        """
        To stop the timer, if it is periodic or single shot.
        """
        if self._timer:
            self._timer.cancel()
        if self._timerSingle:
            self._timerSingle.cancel()


# vim: set tw=79 :



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os.path as osp


class Paths(object):
    """
    Class to manage paths from a given root directory. It expects to have a
    root directory, with a config directory containing configuration files, a
    data directory containing some data resources and a scripts directory
    containing scripts to be executed. This is the standard structure of the
    ROC pipeline.
    """
    def __init__(self, root):
        self.root = root

    def from_root(self, *args):
        """
        Given a list of paths, do a concatenation of them with the root
        directory.
        """
        return osp.join(self._root, *args)

    def from_json_schemas(self, *args):
        """
        Given a list of paths, do a concatenation of them with the json schemas
        directory.
        """
        return osp.join(self.json_schemas, *args)

    def from_config(self, *args):
        """
        Given a list of paths, do a concatenation of them with the configuration
        directory.
        """
        return osp.join(self.config, *args)

    def from_scripts(self, *args):
        """
        Given a list of paths, do a concatenation of them with the scripts
        directory.
        """
        return osp.join(self.scripts, *args)

    def from_data(self, *args):
        """
        Given a list of paths, do a concatenation of them with the data
        directory.
        """
        return osp.join(self.data, *args)

    def from_templates(self, *args):
        """
        Given a list of paths, do a concatenation of them with the templates
        directory.
        """
        return osp.join(self.templates, *args)

    @property
    def root(self):
        return self._root

    @root.setter
    def root(self, root):
        """
        Check that the root is a valid directory.
        """
        if not osp.isdir(root):
            raise ValueError(
                "Root directory {0} is not a valid directory".format(root)
            )

        # store the root
        self._root = root

        # set also the scripts, config and data directories
        self.data = osp.abspath(osp.join(self._root, "data"))
        self.json_schemas = osp.abspath(osp.join(self._root, "json_schemas"))
        self.config = osp.abspath(osp.join(self._root, "config"))
        self.scripts = osp.abspath(osp.join(self._root, "scripts"))
        self.templates = osp.abspath(osp.join(self._root, "templates"))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


"""
Contains Lock class.
Can be used to generate a temporary file in the output directory.
This file is created at the beginning of the run and automatically deleted at the end.

This functionality can be used to preserve from processes overlapping, in order to indicate
current process is still running.
(e.g., moving files that are being written).
"""

import os
from pathlib import Path

from poppy.core.logger import logger

__all__ = ['LockFile']

class LockFile:

    def __init__(self,
                 lock_filename=None,
                 lock_dirpath=None):

        self._started = False

        self.dirpath = lock_dirpath
        self.filename = lock_filename

    @property
    def dirpath(self):
        return self._dirpath

    @dirpath.setter
    def dirpath(self, value):
        self._dirpath = value

    @property
    def filename(self):
        return self._filename

    @filename.setter
    def filename(self, value):
        self._filename = value

    def start(self):

        filepath = self.path()
        if self._started:
            logger.debug('Locking has been already started')
        elif not filepath:
            logger.debug('Locking not started: lock filepath not fully defined!')
            self._started = False
        else:
            try:
                # Make sure that lock file directory exists
                os.makedirs(self.dirpath, exist_ok=True)
                Path(filepath).touch()
            except FileExistsError:
                logger.warning(f'{filepath} already exists and cannot be used as a lock file!')
            else:
                self._started = True
                logger.debug(f'Locking started ({filepath} created)')

    def stop(self):

        # If locking has been started (i.e., started = True)
        # and lock file exists
        # then delete it and stop the locking (i.e., started = False)
        if self._started:
            filepath = self.path()
            if filepath and os.path.isfile(filepath):
                os.remove(filepath)

            self._started = False
            logger.debug(f'Locking stopped ({filepath} removed)')

    def path(self):
        if self.dirpath and self.filename:
            return os.path.join(self.dirpath, self.filename)
        else:
            return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import collections
import os.path as osp
from contextlib import contextmanager

from poppy.core.conf import settings
from poppy.core.db.connector import Connector
from poppy.core.db.dry_runner import DryRunner
from poppy.core.db.handlers import link_databases
from poppy.core.generic.cache import CachedProperty
from poppy.core.generic.graph import Graph
from poppy.core.tools.lock import LockFile
from poppy.core.logger import logger
from poppy.core.properties import Properties
from poppy.core.tools.exceptions import print_exception
from poppy.core.loop import Loop
from poppy.pop.models.job import JobException, exception_level_list, exception_type_list

__all__ = ["Pipeline"]


class PipelineError(Exception):
    """
    Errors for the pipeline.
    """


class EmptyLoopError(Exception):
    """
    Errors for the pipeline loop.
    """


class Pipeline(object):
    """
    Pipeline is the class to manage the task workflow. The module pop is
    responsible for a few more jobs to done, such as scripts to load the
    pipeline and the various commands allowed to test and launch the pipeline.

    All created tasks inside commands will be managed by the Pipeline class.
    """

    def __init__(self, args):
        """
        Initialization of some parameters used by the pipeline.
        """
        # create a property object to be able to share information across
        # tasks
        self.properties = Properties()
        self.properties += vars(args)

        # store the pipeline args
        self.args = args

        # store the pipeline targets
        self.targets = {}

        # store the pipeline configuration
        self.configuration = None

        # link databases
        self._link_databases()

        # set the connector for the POPPy database
        self.db = Connector.manager[settings.MAIN_DATABASE]

        # create the dry runner, which allow to not execute decorated functions
        self.dry_runner = DryRunner()
        self.dry_run = args.dry_run

        # a dict of tasks linked to the pipeline
        # the dict is indexed using the task instance names
        self.tasks = {}

        # a graph containing the topology of the pipeline
        self.graph = Graph()

        # default value of the task used as entry point for the pipeline
        self._entry_point = None
        self._start = None
        self._end = None

        # Initialize ignored_target list
        self.properties.ignored_target = []

        # no new topology to generate by default
        self._is_new_topology = False

        # container for loops between tasks
        self._loops = []

        # Initialize Lock instance
        self.lock = LockFile()

        # Exit attribute (if set to True, then force the exit of the pipeline
        # properly)
        self._exit = False

        # Flag about to set to True if pipeline has no entry point
        self.no_entry_point = False

    @property
    def lock_path(self):
        return self.lock

    @lock_path.setter
    def lock_path(self, dirpath_and_filename):
        if (dirpath_and_filename and
                len(list(dirpath_and_filename)) == 2):
            self.lock.dirpath = dirpath_and_filename[0]
            self.lock.filename = dirpath_and_filename[1]

    @CachedProperty
    def output(self):
        """
        Returns the output path, from the command line, and if not present,
        from the configuration file.
        """
        # check if output defined in command line
        if self.properties.output is not None:
            # validate the directory
            output = osp.expanduser(self.properties.output)
            # self._validate_dir(output)
            return output

        # else check in the configuration file
        path = str(self.properties.configuration["pipeline.output_path"])

        if path.startswith('$ROOT_DIRECTORY'):
            path = path.replace('$ROOT_DIRECTORY', settings.ROOT_DIRECTORY)

        output = osp.expanduser(path)

        # validate directory
        # self._validate_dir(output)
        return output

    @CachedProperty
    def provider(self):
        """
        Returns the provider of this pipeline. First read command line
        arguments, and if not present, uses the value given in the
        configuration file.
        """
        # check if provider defined in command line
        if self.properties.provider is not None:
            provider = self.properties.provider
        else:
            # else check in the configuration file
            provider = self.properties.configuration["pipeline.provider"]

        # return the provider, stored as an attribute
        return provider

    def exit(self):
        """If called, then set self._exit attribute to True."""
        self._exit = True

    def _validate_dir(self, dirname):
        """
        Check if the directory is existing.
        """
        # check directory is existing
        if osp.isdir(dirname):
            return

        # raise error
        raise PipelineError(
            "{0} is not a valid output directory".format(dirname))

    def _link_databases(self):
        """
        Link databases defined in the configuration file to the connectors to
        made them accessible from the registry.
        """
        # check that a configuration file is available
        if "configuration" not in self.properties:
            return

        # get the defined databases
        databases = self.properties.configuration["pipeline.databases"]
        
        # link databases with their connectors
        link_databases(databases)

        # for each defined database of the pipeline, give it the properties
        # (parameters) of the pipeline
        for database in databases:
            Connector.manager[database["identifier"]
                              ].configuration = self.properties

    def run(self):
        """
        Run the tasks registered into the pipeline.
        """

        # Check self.exit value, and exit if true
        if self._exit:
            logger.info('Exiting current pipeline session...')
            return

        # get the entry point of the topology of the pipeline
        entry_point = self.start

        # check that the pipeline can be run
        if entry_point is None:
            if not self.no_entry_point:
                logger.error(
                    "The pipeline is not linked to an entry point task, " +
                    "and cannot be run."
                )
            else:
                logger.debug('Pipeline has not entry point, exiting')
            return

        # generate the new topology if necessary
        self.generate_topology(entry_point)

        # bind the main-db and other connectors
        self.db.bind()

        # run the entry point task, allowing to loop through the graph
        self._run_tasks(entry_point)

    def generate_topology(self, task):
        """
        Generate the dependency graph of the pipeline from the entry point of
        the task.
        """
        # if topology needs to be generated
        if not self._is_new_topology:
            return

        # clear the old graph
        self.graph = Graph()

        # set the pipeline and dependency graph for the entry point task
        self._set_pipeline(task)

        # change the flag to indicate no new generation
        self._is_new_topology = False

    def _set_pipeline(self, task):
        # if the task is already in the graph, we do not need to continue
        if task in self.graph:
            return

        # add the pipeline to the task
        task.pipeline = self

        # instantiate the targets
        # note: we need the pipeline ref to instantiate the targets
        task.instantiate_targets()

        # # FIXME remove legacy target support
        # # add a reference to each output target in the pipeline
        # for target in task.outputs.values():
        #     self.properties[target.id] = target

        # add a reference to the task in the pipeline
        self.tasks[task.instance_name] = task

        # loop over children
        for child in task.descendants():
            # add the child to the graph
            self.graph.add_node(task, child)
            self._set_pipeline(child)

        # do the same for parents
        for parent in task.ancestors():
            self._set_pipeline(parent)

    def _setup_queue(self, entry_point):
        """
        Set up the task queue

        :param entry_point: the first task of the workflow
        """
        # create a queue to store the tasks to call
        self.queue = collections.deque()

        # set the first task in the queue
        self.queue.append(entry_point)

    def _run_tasks(self, entry_point):
        """
        Run each task of the workflow starting by the entry point task

        :param entry_point: the first task of the workflow
        :return:
        """

        # Start locking
        self.lock.start()

        # set up the task queue
        self._setup_queue(entry_point)

        # loop on elements on the queue while it is not empty
        while (len(self.queue) != 0):
            # take a task from the queue
            task = self.queue.pop()

            # if a task is already completed, pass it, since its children
            # should have been already run, same if an error happened already
            # with him
            if task.completed or task.failed:
                continue

            # loop over dependencies of the task
            run_dependency = False
            for dependency in task.ancestors():
                # check the dependency is already completed or not
                if not dependency.completed:
                    # run the task if not completed (append it to the stack)
                    run_dependency = True
                    self.queue.append(dependency)

            # take dependencies from the stack if they need to be run
            if run_dependency:
                continue

            # check that dependencies are completed after all of them have been
            # executed, else a problem occurred in a dependency and the task
            # cannot be run
            for dependency in task.ancestors():
                if not dependency.completed:
                    logger.error(
                        (
                            "Task {0} has a dependency problem " +
                            "with task {1}"
                        ).format(task, dependency)
                    )
                    return

            # if dependencies satisfied, run the task
            try:
                with self._starter(task):
                    self.check_and_run_task(task)
            except RuntimeError:
                # If RunTimeError is "generator didn't yield", then
                # assuming that the exception has been raised due to empty loop
                # generator
                if str(sys.exc_info()[1]) == "generator didn't yield":
                    logger.info("Empty loop")
                elif str(sys.exc_info()[1]) == "generator raised StopIteration":
                    logger.debug("End of the loop")
                else:
                    logger.exception(f'Error running task {task}, skip it!')
            except:
                logger.exception(f'Error running task {task}, skip it!')

            # If exit has been requested in the task then break
            if self._exit:
                logger.info('Exiting the current pipeline session...')
                break

            # then run its child, if completed
            for child in task.descendants():
                # add children to the stack to be used
                self.queue.append(child)

        # Stop locking
        self.lock.stop()

    def check_and_run_task(self, task):
        """
        Responsible to launch the task and check that inputs and outputs are
        correctly created.
        """

        # FIXME: update checks and database links
        # check existence of inputs
        # self._check_existing(task.input())

        # put targets inside the database
        # self._input_target_job_map(task)

        # run the task
        task.run()

        # check existence of outputs
        # self._check_existing(task.output())

        # put outputs inside the database
        # self._output_target_job_map(task)

    def _check_existing(self, targets):
        """
        Check that all the outputs of the given task exist to say that the
        task is completed.
        """
        # loop through target and check existing
        for target in targets:

            # If target is in the ignored_target list,
            # then skip checking
            if target in self.properties.ignored_target:
                logger.debug("Ignore checking of target {0}".format(target))
                continue

            # check target in properties
            if target not in self.properties:
                message = (
                    "The target {0} is not available in the " +
                    "pipeline"
                ).format(target)
                raise PipelineError(message)

            # get the target instance
            tt = self.properties[target]

            # get the target instance
            instance = self.properties[target]

            # not existing, task not completed
            if not instance.is_empty and not instance.exists():
                message = "The target {0} doesn't exist".format(target)
                raise PipelineError(message)

    @contextmanager
    def _starter(self, task):
        """
        Return a context manager for automatic stuff when launching a task such
        setting parameters inside the database, indicating the task status into
        POPPy database, putting I/O files, etc.
        """
        # indicate that the job started
        task.start()

        # launch the task and catch errors occurring while executing
        try:
            yield
        except Exception as exception:
            # record the exception into the database
            self._handle_exception(task, exception)

            # clean if an exception occurred
            task.error()
            return

        # check that all output targets exists before saying that the task is
        # completed
        task.completed = True

        # clean up
        task.stop()

    def _handle_exception(self, task, exception, message=None):
        """
        To create an exception object for the database and store it.
        """
        # print the message of error
        if isinstance(exception, EmptyLoopError):
            # create a specific message for empty loop
            if message is None:
                message = '%s - %s' % (exception.__class__.__name__,
                                       str(exception))
            mess = print_exception(message=message,
                                   log_level='info',
                                   use_traceback=False)
        else:
            mess = print_exception(message)

        # set the exception in the database
        self.create_exception(task, mess)

    @DryRunner.dry_run
    def create_exception(self, task, message,
                         exception_type='ERROR',
                         exception_level='Critical'):
        """
        To create the representation of an exception in the database and store
        it.
        """

        if exception_type not in exception_type_list:
            logger.error(f'Job exception cannot be stored in the database: \n'
                         f'exception type is not valid {exception_type}')
            return None

        if exception_level not in exception_level_list:
            logger.error(f'Job exception cannot be stored in the database: \n'
                         f'exception level is not valid {exception_level}')
            return None

        # create the representation of the exception
        representation = JobException(
            job=task.job(),
            exception_msg=message,
            exception_type=exception_type,
            exception_level=exception_level
        )
        # FIXME type and level may have default value here, but every exception
        # should define its own type and level

        # add into the database
        self.db.add(representation)
        self.db.update_database()

    def __or__(self, task):
        """
        Used to link the first task to the pipeline and be able to loop
        through the task graph to run appropriately the tasks with their
        dependencies.
        """
        # keep a reference to the first task linked to the pipeline, used as an
        # entry point to loop through the graph of dependency
        self.entry_point = task

        # flag to indicate that a new topology of the pipeline has been set and
        # that is necessary to regenerate the dependency graph if running the
        # pipeline
        self._is_new_topology = True

        # return the task
        return task

    def loop(self, start, end, generator):
        """
        Create a loop between two task in the pipeline topology, according to
        the values given by the provided generator.

        Starts by checking if the topology of the pipeline is existing. Then
        regenerate the topology by security. An instance of the loop is created
        that will override the descendants and ancestors of the start and end
        tasks provided in argument, at the appropriate times.
        """
        # the generation of a loop requires a topology for the pipeline, i.e.
        # that the dependency graph has been created, so checks the basics for
        # its generation
        entry_point = self.start

        # check that the pipeline can be run
        if entry_point is None:
            logger.error(
                "The pipeline doesn't have a topology created and cannot " +
                "generate a loop between tasks {0} and {1}".format(
                    start,
                    end,
                )
            )
            return

        # generate the new topology if necessary
        self.generate_topology(entry_point)

        # create the loop
        loop = Loop(self, start, end, generator)

        # add the loop into the container
        self._loops.append(loop)

    @property
    def dry_run(self):
        """
        To change the state of the dry run mode for the pipeline.

        If True, the dry run mode is activated, else it is deactivated.

        :getter: the state of the dry run mode in the pipeline (True/False).
        :setter: the state of the dry run mode in the pipeline (True/False).
        :type: bool
        """
        return self._dry_run

    @dry_run.setter
    def dry_run(self, dry_run):
        self._dry_run = dry_run
        if dry_run:
            self.dry_runner.activate()
        else:
            self.dry_runner.deactivate()

    @property
    def entry_point(self):
        return self._entry_point

    @entry_point.setter
    def entry_point(self, task):
        self._entry_point = task

    @property
    def start(self):
        """
        The starting task, if different from the start of the task chain.

        If the starting task is changing and a task has already been set as the
        starting one, the old task ancestors are restored to the old state.

        The ancestors of the given starting task are kept to be restored
        later if necessary. Then the generator of the task ancestors is
        replaced by a null generator and the new starting task is set to the
        given one.

        :getter: the selected task for the start.
        :setter: the task in the chain that will mark the start of the pipeline.
        :type: :class:`poppy.pop.task.Task`
        """
        return self._start

    @start.setter
    def start(self, task):
        # store the old method for ancestors
        if task is not self._start:
            # if not the first setting
            if self._start is not None:
                # give old method to the task
                self._start.ancestors = self._old_start_ancestors

            # keep a trace of the method of ancestors of the new starting task
            self._old_start_ancestors = task.ancestors

            def null_generator(*args, **kwargs):
                """
                Not doing yield from () because of stupidity of pylint marking
                it as invalid syntax because of python2 for youcompleteme in
                vim.
                """
                return
                yield

            # set the new ancestors for the starting task
            task.ancestors = null_generator

            # set the new starting task
            self._start = task

    @property
    def end(self):
        """
        The ending task, if different from the end of the task chain.

        If the ending task is changing and a task has already been set as the
        ending one, the old task descendants are restored to the old state.

        The descendants of the given ending task are kept to be restored
        later if necessary. Then the generator of the task descendants is
        replaced by a null generator and the new ending task is set to the
        given one.

        :getter: the selected task for the end.
        :setter: the task in the chain that will mark the end of the pipeline.
        :type: :class:`poppy.pop.task.Task`
        """
        return self._end

    @end.setter
    def end(self, task):
        # store the old method for descendants
        if task is not self._end:
            # if not the first setting
            if self._end is not None:
                # give old method to the task
                self._end.descendants = self._old_end_descendants

            # keep a trace of the method of descendants of the new ending task
            self._old_end_descendants = task.descendants

            def null_generator(*args, **kwargs):
                """
                Not doing yield from () because of stupidity of pylint marking
                it as invalid syntax because of python2 for youcompleteme in
                vim.
                """
                return
                yield

            # set the new descendants for the task
            task.descendants = null_generator

            # set the new ending task
            self._end = task

    def get(self, keyword,
            default=None,
            args=False,
            create=False):
        """
        Return a pipeline property.

        :param keyword: property keyword
        :param default: default value if property keyword not found
        :param args: If True, property is an argument
        :param create: If True, create property in pipeline if not found
        :return: property value
        """

        if args:
            property_dict = self.args.__dict__
        else:
            property_dict = self.properties

        # Get value from property, if not found or NoneType use default value
        value = property_dict.get(keyword)

        if value is None and default is not None:
            value = default

            # If not found and create is True, then create property with
            # default value in the pipeline
            if create:
                property_dict[keyword] = default

        return value



import inspect
import logging
import sys

__all__ = ["logger"]

class PoppyLogger:
    @property
    def logger(self):
        calling_module = inspect.currentframe().f_back
        calling_module_name = calling_module.f_globals['__name__']
        return logging.getLogger(calling_module_name)

sys.modules[__name__] = PoppyLogger()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from poppy.core.logger import logger

from functools import wraps
from poppy.core.generic.metaclasses import Singleton

__all__ = ["DryRunner"]




class DryRunnerMeta(Singleton):
    """
    Override the singleton metaclass to initialize some parameters.
    """
    def __init__(cls, *args, **kwargs):
        # init the class as usual
        super(DryRunnerMeta, cls).__init__(*args, **kwargs)

        # set it not activated by default for all derived classes
        cls.activated = False


class DryRunner(object, metaclass=DryRunnerMeta):
    """
    A class to manage methods to be dry run.
    """
    @classmethod
    def dry_run(cls, func):
        """
        A decorator to put a method or function in dry run mode according to
        the parameter stored into the singleton.
        """
        # define the wrapper function
        @wraps(func)
        def wrapper(*args, **kwargs):
            # if not dry run
            if not cls.activated:
                # call effectively the function
                return func(*args, **kwargs)
            # else do nothing
            return None

        return wrapper

    @classmethod
    def activate(cls):
        """
        Activate the dry run mode.
        """
        logger.debug("Activating dry run")
        cls.activated = True

    @classmethod
    def deactivate(cls):
        """
        Deactivate the dry run mode.
        """
        logger.debug("Deactivating dry run")
        cls.activated = False


# vim: set tw=79 :



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__all__ = ['settings']

from poppy.core.generic.metaclasses import Singleton
from . import default_settings as default_settings


class Settings(object, metaclass=Singleton):

    def __init__(self):
        # since this class is a singleton, the init method will only be called once
        self.configure(default_settings)

    def configure(self, settings_obj):
        """
        Update the settings using a settings module or dict
        """
        if isinstance(settings_obj, dict):
            self.from_dict(settings_obj)
        else:
            self.from_module(settings_obj)

    def from_module(self, settings_module):
        # update this dict from settings module (but only for ALL_CAPS settings)
        for setting in dir(settings_module):
            if setting.isupper():
                setattr(self, setting, getattr(settings_module, setting))

    def from_dict(self, settings_dict):
        # update this dict from settings module (but only for ALL_CAPS settings)
        for key, setting in settings_dict.items():
            if key.isupper():
                setattr(self, key, setting)
            else:
                raise Exception(f'The keys of a settings dictionary have to be in upper case (found: {key})')


settings = Settings()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os.path as osp
from poppy.core.logger import logger
import sys

from xml.dom import minidom
from xml.etree import ElementTree
from .xmltodict import xmltodict as xml

__all__ = ["loadXML", "prettify"]




def loadXML(filename, *args, **kwargs):
    """
    To load a file in xml format.
    """
    # check file existence
    if not osp.isfile(filename):
        logger.error(
            "File not found: {0}!".format(filename)
        )
        sys.exit(-1)

    # read the file and parse the xml structure to a dictionary
    logger.info("Loading {0}...".format(filename))
    with open(filename, 'r') as f:
        config = xml.parse(f.read(), *args, **kwargs)

    return config


def prettify(element, indent="\t", encoding="utf-8"):
    """
    To make a nice print of an element of an XML tree.
    """
    rough_string = ElementTree.tostring(element, encoding)
    reparsed = minidom.parseString(rough_string)
    return reparsed.toprettyxml(indent=indent)


# vim: set tw=79 :



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from .manager import Manager

__all__ = [
    "Singleton", "SingletonManager", "ManagerMeta"
]


class Singleton(type):
    """
    A metaclass to create singletons, i.e classes that can have at most only
    one instance created at a given time.
    """
    def __call__(cls, *args, **kwargs):
        """
        Check that an instance is already stored before creating a new one.
        """
        if hasattr(cls, "instance"):
            return cls.instance

        cls.instance = super(Singleton, cls).__call__(*args, **kwargs)

        return cls.instance


class ManagerMeta(type):
    """
    A metaclass to allow simple managing of instances.
    """
    def __new__(cls, name, bases, attr, **kwargs):
        """
        Overridden to remove the keyword arguments coming from the class
        definition, after the metaclass keyword. They can be retrieved as
        keyword arguments in the metaclass provided for the class construction,
        but the default metaclass (type) doesn't handle keyword arguments, so
        we need to remove them before sending it to type.
        """
        return super(ManagerMeta, cls).__new__(cls, name, bases, attr)

    def __init__(
        cls,
        name,
        bases,
        attr,
        manager=Manager,
        after_creation=lambda x, y: None,
    ):
        """
        To initiate the manager of connectors.
        """
        # init the class as usual
        super(ManagerMeta, cls).__init__(name, bases, attr)

        # add a manager of the connectors
        if not hasattr(cls, "manager"):
            # check if a manager is given at the construction of the class
            cls.manager = manager()
            return

        # do something after the class creation
        after_creation(cls, name)

        # also add the class to the manager
        cls.manager.add(name, cls)

    def __call__(cls, *args, **kwargs):
        """
        Register the database when the instance of the database is created.
        """
        # create the instance
        instance = super(ManagerMeta, cls).__call__(*args, **kwargs)

        # register the database
        cls.manager.create(instance)

        return instance


class SingletonManager(ManagerMeta):
    """
    Metaclass to manage instances created with a given name as singletons.
    """
    def __call__(cls, name, *args, **kwargs):
        """
        Check instance already created, else create one and register.
        """
        # check if the instance already exists
        if name in cls.manager:
            return cls.manager[name]

        # create the instance as usual and register it
        return super(SingletonManager, cls).__call__(
            name,
            *args,
            **kwargs
        )

    def __iter__(cls):
        return iter(cls.manager)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from poppy.core.logger import logger

from poppy.core.generic.dot_dict import SimpleDotDict

__all__ = ["Properties"]




class Properties(SimpleDotDict):
    """
    Class for storing parameters used in the pipeline. Tasks in the pipeline
    should share parameters and targets through this dictionary like.
    """
    def __init__(self, *args, **kwargs):
        # init as usual
        super(Properties, self).__init__(*args, **kwargs)

    def __getattr__(self, item):
        return self.__getitem__(item)

    def __setattr__(self, item, value):
        return self.__setitem__(item, value)

    def __add__(self, other):
        """
        To be able to merge two properties by adding them to each other.
        """
        self.update(other)
        return self

    def __radd__(self, other):
        """
        Same as __add__ but in the other order. The properties at most right in
        the merging is preponderant.
        """
        other.update(self)
        return other

    def __iadd__(self, other):
        """
        Incrementation.
        """
        self.update(other)
        return self

    def to_bag(self):
        """
        Transform the data inside the properties into a bag instance containing
        the top level properties.
        """
        class Bag:
            pass

        bag = Bag()

        for key, value in self.items():
            setattr(bag, key, value)

        return bag



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import collections

from poppy.core.logger import logger

__all__ = ["Loop", "LegacyLoop"]


class LoopMetaclass(type):
    """
    Simply keep a reference of created loops, since weirdly their reference is
    not kept inside a calling function with no assignment.
    """

    def __init__(cls, *args, **kwargs):
        """
        Containers for registering instances.
        """
        # create has usual
        super(LoopMetaclass, cls).__init__(*args, **kwargs)

        # add container if not present
        if not hasattr(cls, "instances"):
            cls.instances = []

    def __call__(cls, *args, **kwargs):
        """
        Register the instances.
        """
        # create the instance as usual
        instance = super(LoopMetaclass, cls).__call__(*args, **kwargs)

        # register the instance
        cls.instances.append(instance)

        # return
        return instance


class LegacyLoop(object, metaclass=LoopMetaclass):
    """
    A class to create a loop between two tasks in the pipeline easily.
    """

    def __init__(self, pipeline, start, end, iterable):
        """
        Construct the class with information on the looping tasks in the
        pipeline, and on what to iterate.
        """
        # container to track the monkey patched methods
        self._tracker = collections.defaultdict(collections.deque)

        # indicate if an error occurred in the previous iteration
        self._current_task = None

        # indicate if the loop is running
        self._running = False

        # store start and end loop tasks
        self.pipeline = pipeline
        self.start = start
        self.end = end
        self.iterable = iterable

        # now init the loop
        self.init()

    def init(self):
        """
        Init some states for the loop for all tasks belonging to it.
        """
        # search all paths between the two nodes
        self.paths = self.pipeline.graph.find_all_paths(self.start, self.end)

        # for all tasks, connect to the signal of an error
        for tasks in self.paths:
            for task in tasks:
                # on error, change the topology
                task.errored.connect(self._to_next)

    def reset(self):
        """
        Make a reset of the tasks inside the loop.
        """
        # loop over tasks to reset them
        for tasks in self.paths:
            # this a list of paths, so second loop
            for task in tasks:
                # reset the task
                task.reset()

    def _to_next(self, task):
        """
        To handle what to do when we need to go to a next iteration. The next
        iteration is the whole process of going from the starting task and the
        ending task.
        """
        # change the topology of the loop to reset
        with self.update_context(task):
            # pass to the next iteration
            self.next()

            # change topology of the pipeline
            self._change_next_to_start(task)

            # reset tasks
            self.reset()

            # keep a trace of the current task
            self._current_task = task

    def _change_next_to_start(self, task):
        """
        To change the next task returned by the task.
        """
        # store the old method of the task
        self._tracker[task].append(task.descendants)

        # monkey path the descendants generator
        task.descendants = self._start_generator

    def _start_generator(self):
        """
        Generator to return the start task every time it is called.
        """
        logger.debug("Putting task {0} as start for loop".format(self.start))
        yield self.start

    def update_context(self, task):
        class UpdateContext(object):

            def __enter__(self):
                pass

            def __exit__(instance, type, value, traceback):
                if isinstance(value, StopIteration):
                    # end of the iterator, de-hook the end task
                    logger.debug("End of the loop")

                    # disconnect start if required
                    self.start.started.disconnect(self.starting)

                    # if a task have an error, descendants are those of the end
                    task.descendants = self.end.descendants

        return UpdateContext()

    def starting(self, start):
        """
        To make a first iteration on the iterable before starting the start
        task.
        """
        # call the first iteration to be able to set inputs if needed
        if not self._running:
            self._running = True
            # change the topology of the loop to reset
            with self.update_context(start):
                # pass to the next iteration
                self.next()

        # put the old method to the task
        if self._current_task is not None:
            task = self._current_task
            task.descendants = self._tracker[task].pop()
            task = None

    def next(self):
        """
        To pass to the next iteration of the loop.
        """
        # call the iterator again
        next(self.iterable)

    def delete(self):
        """
        Remove the reference to the loop.
        """
        self.instances.remove(self)

    @property
    def pipeline(self):
        return self._pipeline

    @pipeline.setter
    def pipeline(self, pipeline):
        self._pipeline = pipeline

    @property
    def start(self):
        return self._start

    @start.setter
    def start(self, start):
        self._start = start

        # connect to the started signal to make a first iteration
        self._start.started.connect(self.starting)

    @property
    def end(self):
        return self._end

    @end.setter
    def end(self, end):
        self._end = end

        # connect to the end of the task to do make the loop
        self._end.ended.connect(self._to_next)


class Loop(object, metaclass=LoopMetaclass):
    """
    A class to create a loop between two tasks in the pipeline easily.
    """

    def __init__(self, pipeline, start, end, generator):
        """
        Construct the class with information on the looping tasks in the
        pipeline, and on what to iterate.
        """
        # container to track the monkey patched methods
        self._tracker = collections.defaultdict(collections.deque)

        # indicate if an error occurred in the previous iteration
        self._current_task = None

        # indicate if the loop is running
        self._running = False

        # store start and end loop tasks
        self.pipeline = pipeline
        self.start = start
        self.end = end
        self._generator = generator
        self._iterable = None

        # now init the loop
        self.init()

    @property
    def iterable(self):
        if self._iterable is None:
            self._iterable = self._generator(self)
        return self._iterable

    def init(self):
        """
        Init some states for the loop for all tasks belonging to it.
        """
        # search all paths between the two nodes
        self.paths = self.pipeline.graph.find_all_paths(self.start, self.end)

        # for all tasks, connect to the signal of an error
        for tasks in self.paths:
            for task in tasks:
                # on error, change the topology
                task.errored.connect(self._to_next)

    def reset(self):
        """
        Make a reset of the tasks inside the loop.
        """
        # loop over tasks to reset them
        for tasks in self.paths:
            # this a list of paths, so second loop
            for task in tasks:
                # reset the task
                task.reset()

    def _to_next(self, task):
        """
        To handle what to do when we need to go to a next iteration. The next
        iteration is the whole process of going from the starting task and the
        ending task.
        """
        # change the topology of the loop to reset
        with self.update_context(task):
            # pass to the next iteration
            self.next()

            # change topology of the pipeline
            self._change_next_to_start(task)

            # reset tasks
            self.reset()

            # keep a trace of the current task
            self._current_task = task

    def _change_next_to_start(self, task):
        """
        To change the next task returned by the task.
        """
        # store the old method of the task
        self._tracker[task].append(task.descendants)

        # monkey path the descendants generator
        task.descendants = self._start_generator

    def _start_generator(self):
        """
        Generator to return the start task every time it is called.
        """
        logger.debug("Putting task {0} as start for loop".format(self.start))
        yield self.start

    def update_context(self, task):
        class UpdateContext(object):

            def __enter__(self):
                pass

            def __exit__(instance, type, value, traceback):
                if isinstance(value, StopIteration):
                    # end of the iterator
                    # de-hook the end task and fix dependencies in case of
                    # errors
                    logger.debug("End of the loop")

                    # disconnect start if required
                    self.start.started.disconnect(self.starting)

                    # force descendants/ancestors in case of error during loop tasks processing
                    # in practice, we change the workflow from:
                    # > task_start --> [loop_task_1 -> loop_task_2 -> loop_task_3] --> task_end
                    # to:
                    # > task_start --> task_end
                    # side effect: loops are always considered successful
                    # TODO: allow loops to be considered failed

                    # force the descendants of the loop to be the task(s) just after the loop
                    # 'loop_task_n' is replaced by 'task_end'
                    task.descendants = self.end.descendants

                    for dependency in task.descendants():
                        # force the ancestors task(s) after the loop to depends on the task just before the loop
                        # 'loop_task_3' is replaced by 'task_start'
                        dependency.ancestors = self.start.ancestors

        return UpdateContext()

    def starting(self, start):
        """
        To make a first iteration on the iterable before starting the start
        task.
        """
        # call the first iteration to be able to set inputs if needed
        if not self._running:
            self._running = True
            # change the topology of the loop to reset
            with self.update_context(start):
                # pass to the next iteration
                self.next()

        # put the old method to the task
        if self._current_task is not None:
            task = self._current_task
            task.descendants = self._tracker[task].pop()
            task = None

    def next(self):
        """
        To pass to the next iteration of the loop.
        """
        # call the iterator again
        next(self.iterable)

    def delete(self):
        """
        Remove the reference to the loop.
        """
        self.instances.remove(self)

    @property
    def pipeline(self):
        return self._pipeline

    @pipeline.setter
    def pipeline(self, pipeline):
        self._pipeline = pipeline

    @property
    def start(self):
        return self._start

    @start.setter
    def start(self, start):
        self._start = start

        # connect to the started signal to make a first iteration
        self._start.started.connect(self.starting)

    @property
    def end(self):
        return self._end

    @end.setter
    def end(self, end):
        self._end = end

        # connect to the end of the task to do make the loop
        self._end.ended.connect(self._to_next)

    @property
    def inputs(self):
        _inputs = {}
        for task in self.start.parents:
            _inputs.update(task.outputs)
        return _inputs

    @property
    def outputs(self):
        _outputs = {}
        for task in self.end.children:
            _outputs.update(task.inputs)
        return _outputs



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from urllib.parse import quote_plus

from poppy.core.logger import logger

from contextlib import contextmanager

from sqlalchemy.ext.declarative import DeferredReflection
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.exc import DisconnectionError
from sqlalchemy.exc import OperationalError
from sqlalchemy.exc import ProgrammingError
from sqlalchemy.exc import NoSuchTableError
from sqlalchemy.engine.url import make_url
from sqlalchemy.orm import scoped_session
from sqlalchemy.orm import sessionmaker
from sqlalchemy import create_engine
from sqlalchemy.pool import Pool
from sqlalchemy import event
from functools import wraps
from copy import copy
from poppy.core.tools.exceptions import print_exception
from poppy.core.generic.metaclasses import ManagerMeta
from poppy.core.generic.cache import CachedProperty
from poppy.core.generic.manager import Manager
from poppy.core.generic.signals import Signal

__all__ = ["Database", "DatabaseException"]


@event.listens_for(Pool, "checkout")
def ping_connection(dbapi_connection, connection_record, connection_proxy):
    """
    This is for handling the case of a closed connection for the MySQL
    database, only happening with the AIT and maybe CNES database. This is
    making me crazy, since all standard for session handling in SQLAlchemy seem
    to be respected. This is clearly an hack, whose the problem must be solved
    in a better way.
    """
    cursor = dbapi_connection.cursor()
    try:
        cursor.execute("SELECT 1")
    except:
        # optional - dispose the whole pool
        # instead of invalidating one at a time
        # connection_proxy._pool.dispose()

        # raise DisconnectionError - pool will try
        # connecting again up to three times before raising.
        logger.warning("Connection error, trying again")
        raise DisconnectionError()
    cursor.close()


class DatabaseException(Exception):
    pass


class BaseManager(Manager):
    """
    Special manager for the bases of mapping classes for tables in sqlachemy in
    order to not have problems of 'race conditions' for the creation of the
    first base, if used in several plugins, modules.
    """

    def get(self, name):
        """
        Add a base class with name if not already in the manager, else create a
        base and add it to the manager. The base will be shared across modules
        once we want to get it.
        """
        # if present return it
        if name in self.availables:
            return self.availables[name]

        # else create it and add it
        Base = declarative_base()
        self.add(name, Base)

        return Base


class DatabaseMeta(ManagerMeta):
    """
    Override some behaviour of the manager metaclass to add specific ones.
    """

    def __init__(cls, name, bases, attr):
        # init the class as usual
        super(ManagerMeta, cls).__init__(name, bases, attr)

        # add a manager of the connectors
        if not hasattr(cls, "manager"):
            # manager for database instances
            cls.manager = Manager()

            # also a manager for bases for sqlalchemy
            cls.bases_manager = BaseManager()


class Database(Signal, metaclass=DatabaseMeta):
    """
    A class to manage the connection status to the database and inform all the
    other controllers connected to this class that the connection appeared or
    is gone.
    """

    def __init__(self, name, *args, **kwargs):
        super(Database, self).__init__(*args, **kwargs)

        # set the name
        self.name = name

        # a configuration parameter
        self.binded = False

        # keep the session of the database
        self.session_factory = sessionmaker()

    def connectDatabase(self):
        """
        To make a connection to the database through SQLAlchemy with the
        parameters in the configuration file.
        """
        # check if the database is available
        connected = self.is_available()

        # emit the signal that the database is connected or not, only if change
        if not hasattr(self, "connected") or self.connected is not connected:
            # set the new status
            self.connected = connected

            # make the connections of the database if necessary (reflections,
            # etc)
            self.makeConnections(connected)

            # send the signal with the new status of the database
            self(connected)

    def connectDatabaseAdmin(self):
        """
        To make a connection to the database through SQLAlchemy with the
        parameters in the configuration file.
        """
        # check if the database is available
        connected = self.is_available()

        # emit the signal that the database is connected or not, only if change
        if not hasattr(self, "connected") or self.connected is not connected:
            # set the new status
            self.connected = connected

            # make the connections of the database if necessary (reflections,
            # etc)
            self.makeConnectionsAdmin(connected)

            # send the signal with the new status of the database
            self(connected)

    def makeConnections(self, connected):
        """
        To make the appropriate connections to the database.
        """
        if connected:
            if not self.binded:
                # create the engine of the database, since it is the first time
                # that we have a connection to it
                self.create_engine()

                # reflect the database
                self.reflect()

                # configure the session
                self.configure()

                # store the database connection
                self.create_connection()

                # indicate binded
                self.binded = True

    def makeConnectionsAdmin(self, connected):
        """
        To make the appropriate connections to the database.
        """
        if connected:
            if not self.binded:
                # create the engine of the database, since it is the first time
                # that we have a connection to it
                self.create_engine(admin=True)

                # reflect the database
                self.reflect()

                # configure the session
                self.configure()

                # indicate binded
                self.binded = True

    def unbind(self):
        """
        To indicate that the database is no more binded, allowing to rebind it
        to a new engine.
        """
        self.binded = False

    def configure(self):
        """
        To bind the session to the engine.
        """
        # configure a session
        self.session_factory.configure(bind=self.engine)

    def reflect(self):
        """
        To make the reflection of model classes to the database for the current
        engine, in order to be able to use those classes instances as
        representation of databases.
        """
        # bind the engine to the base class for declarative in order to
        # be able to autoload the structure of the tables directly from
        # the database
        try:
            DeferredReflection.prepare(self.engine)
        except NoSuchTableError as e:
            logger.warning(f'The table {e} does not exist')


    def is_available(self):
        """
        From the package sqlalchemy-utils to check if the database exists, i.e.
        is connected or not.
        """
        # regenerate the url if not already done
        url = self.generate_url()

        # create an url
        url = copy(make_url(url))
        text = 'SELECT 1'
        try:
            # logger.debug("Checking connection to {0}".format(self))
            engine = create_engine(url)
            result = engine.execute(text)
            result.close()
            engine.dispose()
            return True

        except (ProgrammingError, OperationalError):
            logger.exception("Connection error with {0}".format(self))
            return False

    def is_available_with_error(self):
        # regenerate the url if not already done
        url = self.generate_url()

        # create an url
        url = copy(make_url(url))
        text = 'SELECT 1'
        try:
            # logger.debug("Checking connection to {0}".format(self))
            engine = create_engine(url)
            result = engine.execute(text)
            result.close()
            engine.dispose()
            return True

        except (ProgrammingError, OperationalError) as e:
            logger.error("{0}".format(e))
            return False

    def generate_url(self):
        """
        Generate the URL of the database from the parameters.
        """

        vendor = self.parameters.get('vendor', None)

        if vendor == 'sqlite':
            return "{vendor}:///{address}".format(
                **self.parameters
            )
        else:
            # Make sure that special characters are well passed in URL
            user, password = self.parameters['user'].split(":")
            return "{0}://{1}@{2}/{3}".format(
                vendor, quote_plus(user) + ":" + quote_plus(password),
                self.parameters['address'], self.parameters['database']
            )

    def generate_url_admin(self):
        """
        Generate the URL of the database from the parameters, using database
        admin
        """
        if 'admin' not in self.parameters:
            raise KeyError('You need admin credentials to use admin functions')

        vendor = self.parameters.get('vendor', None)

        if vendor == 'sqlite':
            return "{vendor}:///{address}".format(
                **self.parameters
            )
        else:
            # Make sure that special characters are well passed in URL
            admin, password = self.parameters['user'].split(":")
            return "{0}://{1}@{2}/{3}".format(
                vendor, quote_plus(admin) + ":" + quote_plus(password),
                self.parameters['address'], self.parameters['database']
            )


    @classmethod
    def connection(cls, name):
        """
        A decorator in order to provide a wrapper connector to a given
        database, not polluting too much the code with redundant instructions.
        """

        # create a decorator to connect to the database with the given name
        def decorator(func):
            # the wrapper function
            @wraps(func)
            def wrapper(*args, **kwargs):
                # get the database in argument
                if name not in cls.manager:
                    message = "Database {0} doesn't exist".format(name)
                    logger.error(message)
                    raise DatabaseException(message)
                database = cls.manager[name]

                # make the connection to the database if not already done
                database.connectDatabase()

                # call the function as usual
                func(*args, **kwargs)

            return wrapper

        return decorator

    @classmethod
    def is_connected(cls, name):
        """
        Decorator to ensure that a connection exists before running the it.
        """

        # create a decorator to connect to the database with the given name
        def decorator(func):
            # the wrapper function
            @wraps(func)
            def wrapper(*args, **kwargs):
                # get the database in argument
                if name not in cls.manager:
                    message = "Database {0} doesn't exist".format(name)
                    logger.error(message)
                    raise DatabaseException(message)
                database = cls.manager[name]

                # check the connection to the database
                if database.is_available():
                    # call the function as usual
                    func(*args, **kwargs)
                else:
                    message = "Database {0} is not connected".format(database)
                    logger.error(message)
                    raise DatabaseException(message)

            return wrapper

        return decorator

    def create_connection(self):
        no_engine_exception = DatabaseException('You have to instantiate an engine before you can establish a '
                                                'connection')

        if hasattr(self, 'engine'):
            if self.engine:
                self.db_connection = self.engine.connect()
            else:
                raise no_engine_exception
        else:
            raise no_engine_exception

    def create_engine(self, admin=False):
        """
        Create the engine associated to the url of this database.
        """
        if admin:
            url = self.generate_url_admin()
        else:
            url = self.generate_url()

        self.engine = create_engine(url)

        return self.engine

    @property
    def parameters(self):
        return self._parameters

    @parameters.setter
    def parameters(self, parameters):
        self._parameters = parameters

    @CachedProperty
    def scoped_session(self):
        """
        Return a scoped session. This a simply a register for sessions that
        return always the same session if we want to make one. Allows to not
        transfer the session object across the code, simply using it as a
        singleton.
        """
        # check that a session_factory is created
        if hasattr(self, "session_factory"):
            # not already scoped, create a scoped session
            return scoped_session(self.session_factory)
        else:
            logger.error(
                "Trying to create a scoped session on an not existing " +
                "session. Have you created the database?"
            )
            return None

    @contextmanager
    def query_context(self):
        """
        A context manager to create a session for a query and be able to close
        correctly the session when finished, errors, etc.
        """
        # get a session
        session = self.session_factory()

        # return the session and do stuff in the context with it to query
        # information
        try:
            yield session
        except:
            print_exception()
            raise DatabaseException(
                "Error in query context of {0}".format(self)
            )
        finally:
            session.close()

    def __repr__(self):
        return "Database {0}".format(self.name)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['poppy',
 'poppy.core',
 'poppy.core.conf',
 'poppy.core.db',
 'poppy.core.db.models',
 'poppy.core.generic',
 'poppy.core.management',
 'poppy.core.management.templates.pipeline_template',
 'poppy.core.target',
 'poppy.core.task',
 'poppy.core.tools',
 'poppy.core.tools.appdirs',
 'poppy.core.tools.cdf',
 'poppy.core.tools.poppy_logging',
 'poppy.core.tools.test',
 'poppy.core.tools.xmltodict']

package_data = \
{'': ['*'],
 'poppy.core.management': ['templates/plugin_template/*',
                           'templates/plugin_template/plugin_namespace/*',
                           'templates/plugin_template/plugin_namespace/plugin_name/*',
                           'templates/plugin_template/plugin_namespace/plugin_name/models/*',
                           'templates/plugin_template/plugin_namespace/plugin_name/models/versions/*']}

install_requires = \
['SQLAlchemy>=1.3,<2.0',
 'alembic>=1.4,<2.0',
 'declic>=1.0,<2.0',
 'jsonschema>=4.0,<5.0',
 'psycopg2-binary>=2.8.4,<3.0.0']

entry_points = \
{'console_scripts': ['poppy = poppy.core.management.commands:poppy_cli']}

setup_kwargs = {
    'name': 'poppy-core',
    'version': '0.11.2',
    'description': 'POPPy: Plugin-Oriented Pipeline for Python',
    'long_description': 'poppy-core\n===========\n\n[![pipeline status](https://gitlab.obspm.fr/POPPY/POPPyCore/badges/develop/pipeline.svg)](https://gitlab.obspm.fr/POPPY/POPPyCore/pipelines)\n\nThis directory contains the source code of the POPPY framework core.\n\n## Quickstart\n\nTo install package using [pip](https://pypi.org/project/pip-tools/):\n\n```\npip install poppy-core --extra-index-url https://__token__:<your_personal_token>@gitlab.obspm.fr/api/v4/projects/2052/packages/pypi/simple --trusted-host gitlab.obspm.fr\n```\n\nTo install package from source files:\n\n```\ngit clone https://gitlab.obspm.fr/POPPY/POPPyCore.git poppy-core\n```\n\nThen install using [pip](https://pypi.org/project/pip-tools/):\n\n```\ncd poppy-core\npip install .\n```\n\nor using [poetry](https://python-poetry.org/):\n\n```\ncd poppy-core\npip install poetry\npoetry install\n```\n\n## User guide\n\nSee "POPPY User Manual" for more details.\n\n## License\n\nPOPPY is under CeCILL license.\n\n## Acknowledgement\n\nPOPPY is project developed by the RPW Operations Centre (ROC) team based at LESIA (Meudon, France).\nThe ROC is funded by the Centre National d\'Etudes Spatiale (CNES) in the framework of the European Space Agency (ESA) Solar Orbiter mission.\n\n## Authors\n\nxavier.bonnin@obspm.fr (project manager)\n\n\nHas also contributed in the past: Sonny Lion, Gregoire Duvauchelle, Manuel Duarte\n',
    'author': 'Xavier Bonnin',
    'author_email': 'xavier.bonnin@obspm.fr',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.obspm.fr/POPPy/POPPyCore',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4',
}


setup(**setup_kwargs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import argparse

__all__ = ["argparse"]

# make a monkeypatch (I know it is horrible) of the argparse, because of the
# incompatibility of the module from 3.4.2 to 3.4.3, where the defaults
# argument set on a child subparser are not applied if already present in the
# parent subparser, making impossible the definition of a hierarchy in the
# subcommands. This bug is corrected in 3.4.3 and thus we monkeypatch the
# module to make it behaving the same way.


# monkey patch !
def monkeypatch(self, parser, namespace, values, option_string=None):
    parser_name = values[0]
    arg_strings = values[1:]

    # set the parser name if requested
    if self.dest is not argparse.SUPPRESS:
        setattr(namespace, self.dest, parser_name)

    # select the parser
    try:
        parser = self._name_parser_map[parser_name]
    except KeyError:
        args = {'parser_name': parser_name,
                'choices': ', '.join(self._name_parser_map)}
        msg = _('unknown parser %(parser_name)r (choices: %(choices)s)') % args
        raise argparse.ArgumentError(self, msg)

    # parse all the remaining options into the namespace
    # store any unrecognized options on the object, so that the top
    # level parser can decide what to do with them

    # In case this subparser defines new defaults, we parse them
    # in a new namespace object and then update the original
    # namespace for the relevant parts.
    subnamespace, arg_strings = parser.parse_known_args(arg_strings, None)
    for key, value in vars(subnamespace).items():
        setattr(namespace, key, value)

    if arg_strings:
        vars(namespace).setdefault(argparse._UNRECOGNIZED_ARGS_ATTR, [])
        getattr(namespace, argparse._UNRECOGNIZED_ARGS_ATTR).extend(
            arg_strings
        )

argparse._SubParsersAction.__call__ = monkeypatch




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import abc
from contextlib import contextmanager

from poppy.core.conf import settings
from poppy.core.db.connector import Connector
from poppy.core.db.dry_runner import DryRunner
from poppy.core.generic.cache import CachedProperty
from poppy.core.logger import logger

__all__ = ["BaseTarget", ]


class BaseTargetException(Exception):
    pass


class BaseTargetMeta(type):
    """
    The BaseTargetMeta metaclass allows to retrieve an existing instance of the target (using the 'target_id') and
    update it, or to create a new one if the pipeline does not yet contain the 'target_id'
    """

    def __call__(self, target_id, pipeline, *args, many=False, **kwargs):
        """

        :param self: the target class (BaseTarget, FileTarget, etc.)
        :param target_id: the target id
        :param pipeline: the pipeline instance
        :param args: target args
        :param many: the many flag indicating a multi-target instance
        :param kwargs: target kwargs
        :return:
        """
        if target_id in pipeline.targets:
            target = pipeline.targets[target_id]

            # check the compatibility of the target classes
            if type(target) != self:
                raise BaseTargetException(f"Same target_id '{target_id}' was use with two different target classes "
                                          f"('{type(target)}' and '{self}')")

            # check compatibility for multi-target instances
            if target._is_multi_target != many:
                raise BaseTargetException(f"Error in target definitions for target_id '{target_id}'. The target has "
                                          f"been defined as both multi and simple target. Check the 'many' keyword of "
                                          f"your target declarations")

            target.update(target_id, pipeline, *args, many=many, **kwargs)

            return target
        else:
            target = self.__new__(self, target_id, pipeline, *args, many=many, **kwargs)
            target.__init__(target_id, pipeline, *args, many=many, **kwargs)
            pipeline.targets[target_id] = target
            return target


class BaseTarget(metaclass=BaseTargetMeta):
    """
    Base class for the targets of the tasks, which are the input/output used and
    generated by the task. A wrapper is used to easily define the I/O of each
    task while keeping the job of managing it easy, by hiding the database
    representation, the existence or not of the I/O etc.
    """
    PENDING = "Pending"
    STOPPED = "Terminated"
    STARTED = "InProgress"
    EMPTY = "Empty"
    STATE = {PENDING, STOPPED, STARTED, EMPTY}

    OK = "OK"
    WARNING = "WARNING"
    ERROR = "ERROR"
    STATUS = {OK, WARNING, ERROR}

    class TargetEmpty(Exception):
        """
        Exception to indicate that the target is empty.
        """

    def __iter__(self):
        """
        Generate the target instances for multi-target inputs/outputs (many=True)
        """
        raise NotImplementedError(f"The method '__iter__' of the class '{self.__class__.__name__}' is not implemented")

    def update(self, *args, **kwargs):
        """
        Update the target instance with the given args/kwargs

        The update method should prevents conflict (same args defined multiple time) for multiple target initialization
        (for example with input and output declarations)

        Note: the update method is automatically called during instantiation of already existing target instances
        (instance with the same target_id).

        :return:
        """
        pass

    def __init__(self, target_id, pipeline, *args, many=False, **kwargs):
        """
        Initialize a POPPy target.

        :param target_id: Identifier of the target
        :param target_version: Version of the target (01 by default)
        :param target_file: If target is a file, the path to the file
        """

        # store a flag indicating if the the target represent one or more file
        self._is_multi_target = many

        self._pipeline = pipeline

        # the connection object used to communicate with the database
        self.db = Connector.manager[settings.MAIN_DATABASE]

        self._state = self.PENDING
        self._status = self.OK
        self.is_empty = False

        # store the identifier of the target
        self.id = target_id

        # call the update function to handle merge conflicts
        self.update(*args, **kwargs)

    def link(self, target_id):
        """
        Replace the 'target_id' instance by the current one in the pipeline object

        :param target_id:
        :return:
        """

        self._pipeline.targets[target_id] = self

    def exists(self):
        """
        Returns ``True`` if the target exists and ``False`` otherwise.
        """
        pass

    def state(self, value):
        """
        To change the state of the target with the given provided argument.
        """
        # check that the value is accepted
        if value not in self.STATE:
            raise ValueError(
                "The state of a target must be set to {0}".format(self.STATE)
            )

        # store the state
        self._state = value

        # call internal method to change the state of a target
        self.target_changed()

    def status(self, value):
        """
        To change the status of the target.
        """
        # check that the value is accepted
        if value not in self.STATUS:
            raise ValueError(
                "The status of a target must be set to " +
                "{0}".format(self.STATUS)
            )

        # store the status
        self._status = value

        # indicate the status changed
        self.target_changed()

    @DryRunner.dry_run
    def target_changed(self):
        """
        Callback used when target changed.
        """
        pass

    def ok(self):
        """
        Set to ok the status
        """
        self.status(self.OK)

    def warning(self):
        """
        Set the status to warning.
        """
        self.status(self.WARNING)

    def error(self):
        """
        Set the status to error.
        """
        self.status(self.ERROR)

    def pending(self):
        """
        Set the state to pending.
        """
        self.state(self.PENDING)

    def terminated(self):
        """
        Set the state to terminated.
        """
        self.state(self.STOPPED)

    def progress(self):
        """
        Set the state to in progress.
        """
        self.state(self.STARTED)

    def empty(self):
        """
        Set the state to empty.
        """
        self.is_empty = True
        self.state(self.EMPTY)

    @abc.abstractmethod
    def open(self, *args, **kwargs):
        """
        A wrapper around the open function of the file.
        """
        pass

    @contextmanager
    def activate(self):
        """
        To use the target as a wrapper around other files type that cannot be
        opened as usual.
        """
        # mark as pending and ok
        self.ok()
        self.pending()

        # try to do the things inside the context
        try:
            # mark as in progress
            self.progress()
            yield
        except self.TargetEmpty as e:
            self.empty()
        except Exception as e:
            self.error()
            self.terminated()
            raise e
        else:
            # mark as terminated
            self.terminated()

    @CachedProperty
    def description(self):
        """Given the target id, get the target target description from the database."""
        return self.db.get_target(self.id)

    def __repr__(self):
        return self.id

    @classmethod
    def output(cls, identifier, *args, **kwargs):
        """
        Define a new output target for a given class

        :param identifier: the target identifier
        :return: the class wrapper
        """

        def wrapper(TaskClass):
            TaskClass.add_output(target_class=cls, identifier=identifier, *args, **kwargs)
            return TaskClass

        return wrapper

    @classmethod
    def input(cls, identifier, *args, **kwargs):
        """
        Define a new input target for a given class

        :param identifier: the target identifier
        :return: the class wrapper
        """

        def wrapper(TaskClass):
            TaskClass.add_input(target_class=cls, identifier=identifier, *args, **kwargs)
            return TaskClass

        return wrapper



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__all__ = ["CachedProperty"]


class CachedProperty(object):
    """
    A descriptor to cache the execution of a property, and sending the computed
    value again and again.
    """

    def __init__(self, method):
        """
        Store the wrapped method, and its name.
        """
        self.name = method.__name__
        self.func = method

    def __get__(self, instance, owner):
        """
        When getting the attribute the first time, the descriptor is called and
        then compute the method (no arguments) as usual. Then the method in
        the instance is replaced by an attribute containing the cached value.
        """
        # compute the method as usual
        result = self.func(instance)

        # replace the method by the attribute, avoiding to call the descriptor
        # again !
        setattr(instance, self.name, result)

        # return the result
        return result



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from poppy.core.db.database import Database
from poppy.core.conf import settings

__all__ = ["Base"]

# create a base declarative class for all models
Base = Database.bases_manager.get(settings.MAIN_DATABASE)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import datetime
import os
import os.path as osp
import uuid
from contextlib import contextmanager

from poppy.core.db.dry_runner import DryRunner
from poppy.core.generic.cache import CachedProperty
from poppy.core.logger import logger
from poppy.core.target.base_target import BaseTarget
from poppy.pop.tools import compute_file_sha

__all__ = ["FileTarget"]


class FileTargetException(Exception):
    pass


class FileTarget(BaseTarget):
    """
    FileTarget class
    """

    def __init__(self, *args, filepath=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.filepath = filepath

    def update(self, *args, filepath=None, **kwargs):
        """
        Update the file target instance with the given args/kwargs

        Updated attributes:
          - filepath

        :return:
        """

        super().update(*args, **kwargs)

        if filepath is not None:
            if self.filepath != filepath:  # and self.filepath is not None:
                logger.warning(f"Filepath of FileTarget '{self.id}' was defined multiple times")
            self.filepath = filepath

    def __iter__(self):
        if not self._is_multi_target:
            raise FileTargetException("Simple targets are not iterable, use the 'many=True' flag to generate "
                                      "multi-target objects")

        for idx, filepath in enumerate(self.filepath):
            # TODO: add parametrization of the target id
            yield FileTarget(f"__{self.id}__{idx}__", self._pipeline, filepath=filepath)

    @property
    def filepath(self):
        if self._filepath:
            return self._filepath
        if callable(self._filepath_generator):
            self._filepath = self._filepath_generator(self._pipeline)
            return self._filepath
        return None

    @filepath.setter
    def filepath(self, value):
        # if the filepath was given as a function
        if callable(value):
            self._filepath_generator = value
            self._filepath = None
        else:
            self._filepath_generator = None
            self._filepath = value

    @property
    def file_directory(self):
        if not self._file_directory:
            self._file_directory = osp.dirname(self.filepath)
        return self._file_directory

    @file_directory.setter
    def file_directory(self, value):
        if not osp.isdir(value):
            logger.error("INPUT VALUE IS NOT A VALID DIRECTORY!")
            raise IsADirectoryError
        else:
            self._file_directory = value

        # Update the file path
        self.filepath = osp.join(value, self.filename)

        # indicate that dirname has changed
        self.target_changed()

    @property
    def filename(self):
        if not self._filename:
            self._filename = osp.basename(self.filepath)
        return self._filename

    def get_filename(self):
        return self.name

    def get_filedir(self):
        return self.dir

    @CachedProperty
    def uuid(self):
        # create an uuid
        return uuid.uuid4()

    def get_filesha(self):
        # Return the sha256 of the target file (hexdigest)
        if not self.sha:
            try:
                self.sha = compute_file_sha(self.filepath)
            except:
                logger.debug(
                    "SHA256 cannot be computed because file does not exist, skipping!")
            else:
                self.sha = self.sha.hexdigest()

        return self.sha

    def set_filesha(self):
        """Set the SHA (in case of file update)"""
        try:
            self.sha = compute_file_sha(self.filepath)
        except:
            logger.debug(
                "SHA256 cannot be computed because file does not exist, skipping!")
        else:
            self.sha = self.sha.hexdigest()

        # indicate that filesha has changed
        self.target_changed()

        return self.sha

    def set_dirname(self, value):
        """
        To change the directory of the target file.
        """
        self.file_directory = value

    @DryRunner.dry_run
    def target_changed(self):
        """
        Recreate the representation of the target if it changed.
        """
        return  # FIXME: create/handle new target database model

        # modify the status
        self.representation.file_state = self._state
        self.representation.file_status = self._status

        # modify the file_dir
        self.representation.file_dir = self.dir

        # modify the file_sha
        self.representation.file_sha = self.sha

    def open(self, *args, **kwargs):
        """
        A wrapper around the open function of the file.
        """
        if self._is_multi_target:
            raise FileTargetException("Multi-target instances are not openable, use the 'many=False' flag to generate "
                                      "a simple target object")

        class FileOpen(object):

            def __enter__(instance):
                self.pending()
                self.file = open(self.filepath, *args, **kwargs)
                return self.file

            def __exit__(instance, type, value, traceback):
                if type is not None:
                    self.error()
                self.file.close()
                self.terminated()

        return FileOpen()

    @contextmanager
    def activate(self):
        """
        To use the target as a wrapper around other files type that cannot be
        opened as usual.
        """
        # mark as pending and ok
        self.ok()
        self.pending()

        # try to do the things inside the context
        try:
            # mark as in progress
            self.progress()
            yield
        except self.TargetEmpty as e:
            self.empty()
        except Exception as e:
            self.error()
            self.terminated()
            raise e
        else:
            # mark as terminated
            self.terminated()

    def exists(self):
        """
        Check that the file exists on the system.
        """
        if self.filepath:
            return osp.isfile(self.filepath)
        else:
            # For now, return None not target file
            logger.warning(
                "Target {0} seems to not be a file!".format(self.id))
            return None

    def size(self):
        """
        Return the size in bytes of the file, 0 if the file doesn't exist or
        there is any kind of problem.
        """
        try:
            return os.stat(self.filepath).st_size
        except:
            return 0

    def creation_date(self):
        """
        To get the creation date of a file. ctime does not give the creation
        time but the change time. Use mtime for last modification date instead.
        """
        try:
            return datetime.datetime.fromtimestamp(osp.getmtime(self.filepath))
        except:
            return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import hashlib

def sha256_from_file(filepath, blocksize=4096):
    hasher = hashlib.sha256()
    with open(filepath, "rb") as f:
        # chunk for long files
        for chunk in iter(lambda: f.read(blocksize), b""):
            hasher.update(chunk)
    return hasher.hexdigest()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os.path as osp
import logging
import os
try:
    import pwd
except ImportError:
    import getpass
    pwd = None

__all__ = ['RotatingFileHandlerPerUser']


def current_user():
    if pwd:
        return pwd.getpwuid(os.getuid()).pw_name
    else:
        return getpass.getuser()


class RotatingFileHandlerPerUser(logging.handlers.RotatingFileHandler):
    def __init__(self, filename, *args, **kwargs):
        # modify the name of the file by adding the user name in prefix
        # first get the dirname and the basename
        dirname, basename = osp.dirname(filename), osp.basename(filename)

        # add the user name to the filename, the name of the user is retrieved
        # for various environment variables in all OS, so it is easy to change
        # the user running based on this. This is not a secure way of doing
        # this but in this case it is not critical,
        try:
            username = current_user()
        except:
            username = 'poppy'

        new_filename = osp.join(
            dirname,
            '_'.join([username, basename])
        )

        # init as usual but with new name
        super(RotatingFileHandlerPerUser, self).__init__(
            new_filename,
            *args,
            **kwargs
        )
